import numpy as np

from batchnormlstm.layers import *
from batchnormlstm.layer_utils import *


class TwoLayerLSTM(object):  # Many-to-One
	"""
	A two-layer recurrent neural network with mean squared error loss
	that uses a modular layer design. We assume an input dimension
	of D, a hidden dimension of H, and output dimension of F.

	The architecure should be LSTM - affine.

	Note that this class does not implement gradient descent; instead, it
	will interact with a separate Solver object that is responsible for running
	optimization.

	The learnable parameters of the model are stored in the dictionary
	self.params that maps parameter names to numpy arrays.
	"""

	def __init__(self, input_dim=3 * 32 * 32, hidden_dim=100, output_dim=None,
				 weight_scale=1e-3, reg=0.0):
		"""
		Initialize a new network.

		Inputs:
		- input_dim: An integer giving the size of the input
		- hidden_dim: An integer giving the size of the hidden layer
		- output_dim: An integer giving the size of the output. Set to the
		  same as input_dim if not specified
		- weight_scale: Scalar giving the standard deviation for random
		  initialization of the weights.
		- reg: Scalar giving L2 regularization strength.
		"""
		self.params = {}
		self.reg = reg

		############################################################################
		# Initialize the weights and biases of the two-layer net. Weights    	   #
		# should be initialized from a Gaussian with standard deviation equal to   #
		# weight_scale, and biases should be initialized to zero. All weights and  #
		# biases should be stored in the dictionary self.params, with first layer  #
		# weights and biases using the keys 'W1' and 'b1' and second layer weights #
		# and biases using the keys 'W2' and 'b2'.                                 #
		############################################################################
		output_dim = input_dim if output_dim is None else output_dim

		self.params['W1'] = np.random.normal(0, weight_scale, [input_dim, 4 * hidden_dim])
		self.params['U'] = np.random.normal(0, weight_scale, [hidden_dim, 4 * hidden_dim])
		self.params['b1'] = np.zeros(4 * hidden_dim)
		self.params['W2'] = np.random.normal(0, weight_scale, [hidden_dim, output_dim])
		self.params['b2'] = np.zeros(output_dim)
		############################################################################
		#                            					                           #
		############################################################################

	def loss(self, X, Y=None):
		"""
		Compute loss and gradient for a minibatch of data.

		Inputs:
		- X: Array of input data of shape (N, T, d_1, ..., d_k)
		- Y: Array of targeted vector of shape (N, f_1, ..., f_k).

		Returns:
		If Y is None, then run a test-time forward pass of the model and return:
		- out: Array of shape (N, F) giving regression output, where
		  out[i, j] is the predicted value for the j-th dimension of the i-th data.

		If Y is not None, then run a training-time forward and backward pass and
		return a tuple of:
		- loss: Scalar value giving the loss
		- grads: Dictionary with the same keys as self.params, mapping parameter
		  names to gradients of the loss with respect to those parameters.
		"""
		N, T = np.shape(X)[0:2]
		M, F = np.shape(self.params['W2'])
		h_prev, c_prev = np.zeros((2, N, M))
		out = None
		############################################################################
		# Implement the forward pass for the two-layer net, computing the  		   #
		# regression output for X and storing them in the scores variable.         #
		############################################################################
		lstm_caches = [None] * T
		for t in range(T):
			h, c, lstm_cache = lstm_forward_unit(X[:, t, ...], self.params['W1'],
												 self.params['U'], self.params['b1'],
												 h_prev, c_prev)
			lstm_caches[t] = lstm_cache
			h_prev, c_prev = h, c
		out, affine_cache = affine_forward(h_prev, self.params['W2'], self.params['b2'])
		############################################################################
		#                            					                           #
		############################################################################

		# If Y is None then we are in test mode so just return scores
		if Y is None:
			return out

		loss, grads = 0, {}
		############################################################################
		# Implement the backward pass for the two-layer net. Store the loss  	   #
		# in the loss variable and gradients in the grads dictionary. Compute data #
		# loss using MSE, and make grads[k] hold the gradients for 	  		       #
		# self.params[k]. Then add L2 regularization. L2 regularization includes   #
		# a factor of 0.5 to simplify the expression for the gradient.             #
		############################################################################
		err = out - Y.reshape(N, -1)  # N*F
		MSE = np.mean(err ** 2, axis=1)  # N
		loss = np.mean(MSE) + self.reg * 0.5 * (np.sum(self.params['W1'] ** 2) + np.sum(self.params['W2'] ** 2) +
												np.sum(self.params['U'] ** 2))

		dout = 1. / N * 2. / F * err  # N*F
		dmid, dw2, db2 = affine_backward(dout, affine_cache)
		grads['W2'] = dw2 + self.reg * self.params['W2']  # M*F
		grads['b2'] = db2  # F

		douts = np.concatenate((np.zeros((N, T - 1, M)), dmid[:, np.newaxis, :]), axis=1)  # N*T*M
		dh, dc_next, f_next = np.zeros((3, N, M))
		grads['W1'] = np.zeros_like(self.params['W1'])
		grads['U'] = np.zeros_like(self.params['U'])
		grads['b1'] = np.zeros_like(self.params['b1'])
		for t in reversed(range(T)):
			dx, dw, du, db, dh_prev, dc, f = lstm_backward_unit(douts[:, t, :], lstm_caches[t], dh, dc_next, f_next)
			grads['W1'] += dw  # D*M
			grads['U'] += du  # D*M
			grads['b1'] += db  # M
			dh, dc_next, f_next = dh_prev, dc, f
		grads['W1'] += self.reg * self.params['W1']
		grads['U'] += self.reg * self.params['U']
		############################################################################
		#                             				                               #
		############################################################################

		return loss, grads


class AdvancedLSTM(object):
	"""
	A recurrent neural network with an arbitrary number of hidden LSTM layers
	and a mean squared error loss function. This will also implement
	dropout and batch normalization as options.
	Batch-Normalized LSTM: https://arxiv.org/pdf/1603.09025.pdf
	LSTM with Dropout: https://arxiv.org/pdf/1409.2329.pdf

	For a network with L layers, the architecture will be

	{[dropout] [batch norm] LSTM} x (L - 1) - affine

	where batch normalization and dropout are optional, and the {...} block is
	repeated L - 1 times.

	Similar to the TwoLayerLSTM above, learnable parameters are stored in the
	self.params dictionary and will be learned using the Solver class.
	"""

	def __init__(self, hidden_dims, input_dim=3 * 32 * 32, output_dim=None,
				 dropout=0, use_batchnorm=False, reg=0.0,
				 weight_scale=1e-2, dtype=np.float32, seed=None):
		"""
		Initialize a new AdvancedLSTM.

		Inputs:
		- hidden_dims: A list of integers giving the size of each hidden layer.
		- input_dim: An integer giving the size of the input.
		- output_dim: An integer giving the size of the input. Set to the same
		  as input_dim if not specified.
		- dropout: Scalar between 0 and 1 giving dropout strength. If dropout=0 then
		  the network should not use dropout at all.
		- use_batchnorm: Whether or not the network should use batch normalization.
		- reg: Scalar giving L2 regularization strength.
		- weight_scale: Scalar giving the standard deviation for random
		  initialization of the weights.
		- dtype: A numpy datatype object; all computations will be performed using
		  this datatype. float32 is faster but less accurate, so you should use
		  float64 for numeric gradient checking.
		- seed: If not None, then pass this random seed to the dropout layers. This
		  will make the dropout layers deteriminstic so we can gradient check the
		  model.
		"""
		self.use_batchnorm = use_batchnorm
		self.use_dropout = dropout > 0
		self.reg = reg
		self.num_layers = 1 + len(hidden_dims)
		self.dtype = dtype
		self.params = {}
		self.features = {}

		############################################################################
		# Initialize the parameters of the network, storing all values in   	   #
		# the self.params dictionary. Store weights and biases for the first layer #
		# in W1, U1 and b1; for the second layer use W2, U2 and b2, etc.(for the   #
		# last layer there is no U). Weights should be initialized from a normal   #
		# distribution with standard deviation equal to weight_scale and biases    #
		# should be initialized to zero.         								   #
		#                                                                          #
		# When using batch normalization, store scale and shift parameters for the #
		# first layer in gamma_{.}1 and beta_{.}1; for the second layer use 	   #
		# gamma_{.}2 and beta_{.}2, etc. Scale parameters should be initialized to #
		# one and shift parameters should be initialized to zero.                  #
		############################################################################
		if type(hidden_dims) != list:
			raise ValueError('hidden_dim has to be a list.')

		output_dim = input_dim if output_dim is None else output_dim
		dims = [input_dim] + hidden_dims + [output_dim]

		for i in range(self.num_layers - 1):
			self.params['W' + str(i + 1)] = np.random.normal(0, weight_scale, [dims[i], 4 * dims[i + 1]])
			self.params['U' + str(i + 1)] = np.random.normal(0, weight_scale, [dims[i + 1], 4 * dims[i + 1]])
			self.params['b' + str(i + 1)] = np.zeros(4 * dims[i + 1])

			if self.use_batchnorm:
				self.params['gamma_x' + str(i + 1)] = np.ones(4 * dims[i + 1])
				self.params['beta_x' + str(i + 1)] = np.zeros(4 * dims[i + 1])
				self.params['gamma_h' + str(i + 1)] = np.ones(4 * dims[i + 1])
				self.params['beta_h' + str(i + 1)] = np.zeros(4 * dims[i + 1])
				self.params['gamma_c' + str(i + 1)] = np.ones(dims[i + 1])
				self.params['beta_c' + str(i + 1)] = np.zeros(dims[i + 1])

		self.params['W' + str(self.num_layers)] = np.random.normal(0, weight_scale,
																   [dims[self.num_layers - 1], dims[self.num_layers]])
		self.params['b' + str(self.num_layers)] = np.zeros(dims[self.num_layers])
		############################################################################
		#                              				                               #
		############################################################################

		# When using dropout we need to pass a dropout_param dictionary to each
		# dropout layer so that the layer knows the dropout probability and the mode
		# (train / test). We can pass the same dropout_param to each dropout layer.
		self.dropout_param = {}
		if self.use_dropout:
			self.dropout_param = {'mode': 'train', 'p': dropout}
			if seed is not None:
				self.dropout_param['seed'] = seed

		# With batch normalization we need to keep track of running means and
		# variances, so we need to pass a special bn_param object to each batch
		# normalization layer. We pass self.bn_params[0] to the forward pass of
		# the first batch normalization layer, self.bn_params[1] to the forward
		# pass of the second batch normalization layer, etc.
		self.bn_params = []
		if self.use_batchnorm:
			self.bn_params = [[{'mode': 'train'} for j in range(3)] for i in range(self.num_layers - 1)]

		# Cast all parameters to the correct datatype
		for k, v in self.params.items():
			self.params[k] = v.astype(dtype)

	def loss(self, X, Y=None):
		"""
		Compute loss and gradient for the AdvancedLSTM.

		Input / output: Same as TwoLayerLSTM above.
		"""
		X = X.astype(self.dtype)
		mode = 'test' if Y is None else 'train'

		# Set train/test mode for batchnorm params and dropout param since they
		# behave differently during training and testing.
		if self.dropout_param is not None:
			self.dropout_param['mode'] = mode
		if self.use_batchnorm:
			for bn_params in self.bn_params:
				for bn_param in bn_params:
					bn_param[mode] = mode

		out = None
		############################################################################
		# Implement the forward pass for the AdvancedLSTM, computing the   		   #
		# regression output for X and storing them in the out variable.            #
		#                                                                          #
		# When using dropout, we'll need to pass self.dropout_param to each        #
		# dropout forward pass.                                                    #
		#                                                                          #
		# When using batch normalization, we'll need to pass self.bn_params[0] to  #
		# the forward pass for the first batch normalization layer, pass           #
		# self.bn_params[1] to the forward pass for the second batch normalization #
		# layer, etc.                                                              #
		############################################################################
		dropout_caches = [None] * (self.num_layers - 1)
		lstm_cacheses = [None] * (self.num_layers - 1)
		affine_cache = None

		mids = X
		for i in range(self.num_layers - 1):
			# dropout input
			if self.use_dropout:
				mids, dropout_caches[i] = dropout_forward(mids, self.dropout_param)
			# batchnorm lstm
			if self.use_batchnorm:
				mids, lstm_cacheses[i] = batchnorm_lstm_forward(mids, self.params['W' + str(i + 1)],
																self.params['U' + str(i + 1)],
																self.params['b' + str(i + 1)],
																(self.params['gamma_x' + str(i + 1)],
																 self.params['gamma_h' + str(i + 1)],
																 self.params['gamma_c' + str(i + 1)]),
																(self.params['beta_x' + str(i + 1)],
																 self.params['beta_h' + str(i + 1)],
																 self.params['beta_c' + str(i + 1)]),
																self.bn_params[i])
			else:
				mids, lstm_cacheses[i] = lstm_forward(mids, self.params['W' + str(i + 1)],
													  self.params['U' + str(i + 1)], self.params['b' + str(i + 1)])
			# store feature
			self.features['feature' + str(i + 1)] = mids
		out, affine_cache = affine_forward(mids[:, -1, ...], self.params['W' + str(self.num_layers)],
										   self.params['b' + str(self.num_layers)])
		############################################################################
		#                             				                               #
		############################################################################

		# If test mode return early
		if mode == 'test':
			return out

		loss, grads = 0.0, {}
		############################################################################
		# Implement the backward pass for the fully-connected net. Store the       #
		# loss in the loss variable and gradients in the grads dictionary. Compute #
		# data loss using mse_loss, and make sure that grads[k] holds the 		   #
		# gradients for self.params[k]. Then add L2 regularization. 			   #
		# L2 regularization includes a factor of 0.5 to simplify the expression    #
		# for the gradient.   													   #
		#                                                                          #
		# When using batch normalization, we don't need to regularize the scale    #
		# and shift parameters.                                                    #
		############################################################################
		N, T = np.shape(X)[0:2]
		M, F = np.shape(self.params['W' + str(self.num_layers)])
		loss, dout = mse_loss(out, Y.reshape(N, -1))
		loss += self.reg * 0.5 * sum([np.sum(self.params['W' + str(i + 1)] ** 2) for i in range(self.num_layers)])
		loss += self.reg * 0.5 * sum([np.sum(self.params['U' + str(i + 1)] ** 2) for i in range(self.num_layers - 1)])

		dmid, dw, db = affine_backward(dout, affine_cache)
		grads['W' + str(self.num_layers)] = dw + self.reg * self.params['W' + str(self.num_layers)]  # M*F
		grads['b' + str(self.num_layers)] = db  # F

		dmids = np.concatenate((np.zeros((N, T - 1, M)), dmid[:, np.newaxis, :]), axis=1)  # N*T*M
		for i in reversed(range(self.num_layers - 1)):
			if self.use_batchnorm:
				dmids, dw, du, db, dgammas, dbetas = batchnorm_lstm_backward(dmids, lstm_cacheses[i])
				grads['W' + str(i + 1)] = dw + self.reg * self.params['W' + str(i + 1)]
				grads['U' + str(i + 1)] = du + self.reg * self.params['U' + str(i + 1)]
				grads['b' + str(i + 1)] = db
				dgamma_x, dgamma_h, dgamma_c = dgammas
				dbeta_x, dbeta_h, dbeta_c = dbetas
				grads['gamma_x' + str(i + 1)] = dgamma_x
				grads['gamma_h' + str(i + 1)] = dgamma_h
				grads['gamma_c' + str(i + 1)] = dgamma_c
				grads['beta_x' + str(i + 1)] = dbeta_x
				grads['beta_h' + str(i + 1)] = dbeta_h
				grads['beta_c' + str(i + 1)] = dbeta_c
			else:
				dmids, dw, du, db = lstm_backward(dmids, lstm_cacheses[i])
				grads['W' + str(i + 1)] = dw + self.reg * self.params['W' + str(i + 1)]
				grads['U' + str(i + 1)] = du + self.reg * self.params['U' + str(i + 1)]
				grads['b' + str(i + 1)] = db

			if self.use_dropout:
				dmids = dropout_backward(dmids, dropout_caches[i])
		dx = dmids
		############################################################################
		#                              				                               #
		############################################################################

		return loss, grads

	def save(self, path=".\\model.txt"):
		import json
		with open(path, 'w') as file:
			file.write(json.dumps(self.params))

	def load(self, path=".\\model.txt"):
		import json
		self.params = json.load(open(path))

